use std::time::{Duration, Instant};

use color_eyre::{owo_colors::OwoColorize, Result};
use ratatui::{
	buffer::Buffer,
	crossterm::event::{self, KeyCode, KeyEventKind},
	layout::{Constraint, Layout, Rect},
	style::{Color, Style, Stylize},
	symbols::Marker,
	widgets::{
		canvas::{Canvas, Rectangle, Shape},
		Block, BorderType, Widget,
	},
	DefaultTerminal,
};
use Constraint::{Fill, Length, Percentage};

use crate::gamepad_manager::GamepadManager;

const MAX_DEADZONE: f64 = 0.7;
const MIN_DEADZONE: f64 = 0.6;
const REST_DEADZONE: f64 = 0.05;

pub struct App {
	pub manager: GamepadManager,

	state: AppState,
	tick: u64,
}

#[derive(Default, PartialEq, Eq)]
pub enum AppState {
	#[default]
	Running,
	Quitting,
}

impl App {
	pub fn new(manager: GamepadManager) -> Self {
		App {
			manager,
			state: AppState::Running,

			tick: 0,
		}
	}

	pub fn run(mut self, mut terminal: DefaultTerminal) -> Result<()> {
		let tick_rate = Duration::from_millis(20);
		let mut last_tick = Instant::now();

		while self.state == AppState::Running {
			terminal.draw(|frame| frame.render_widget(&self, frame.area()))?;

			let timeout = tick_rate.saturating_sub(last_tick.elapsed());
			if event::poll(timeout)? {
				self.handle_events()?;
			}
			self.handle_gamepad_inputs()?;

			if last_tick.elapsed() >= tick_rate {
				self.update();
				last_tick = Instant::now();
			}
		}
		Ok(())
	}

	fn handle_gamepad_inputs(&mut self) -> Result<()> {
		while let Some(gilrs::Event { id, .. }) = self.manager.gilrs.next_event() {
			self.manager.set_active_gamepad(id);
		}
		Ok(())
	}

	fn handle_events(&mut self) -> Result<()> {
		if let ratatui::crossterm::event::Event::Key(key) = event::read()? {
			if key.kind == KeyEventKind::Press {
				match key.code {
					KeyCode::Char('r') => {
						self.manager.scan_gamepads();
						return Ok(());
					}
					KeyCode::Char('v') => self.manager.test_weak_ff(),
					KeyCode::Char('V') => self.manager.test_strong_ff(),
					KeyCode::Char('q') | KeyCode::Char('Q') | KeyCode::Esc => self.quit(),
					_ => {}
				}
			}
		}
		Ok(())
	}

	fn update(&mut self) {
		self.tick = self.tick.wrapping_add(1);
	}

	fn quit(&mut self) {
		self.state = AppState::Quitting;
	}

	fn render_action_buttons(&self, area: Rect, buf: &mut Buffer) {
		Block::bordered().render(area, buf);

		let north_button = self.create_button_ui(gilrs::Button::North, "Y");
		let east_button = self.create_button_ui(gilrs::Button::East, "B");
		let south_button = self.create_button_ui(gilrs::Button::South, "A");
		let west_button = self.create_button_ui(gilrs::Button::West, "X");

		let layers = Layout::vertical([Fill(1), Fill(1), Fill(1)]);
		let [top, mid, bot] = layers.areas(area);

		let top_layer = Layout::horizontal([Fill(1), Fill(1), Fill(1)]);
		let [_, north, _] = top_layer.areas(top);

		let mid_layer = Layout::horizontal([Fill(1), Fill(1), Fill(1)]);
		let [west, _, east] = mid_layer.areas(mid);

		let bot_layer = Layout::horizontal([Fill(1), Fill(1), Fill(1)]);
		let [_, south, _] = bot_layer.areas(bot);

		north_button.render(north, buf);
		east_button.render(east, buf);
		south_button.render(south, buf);
		west_button.render(west, buf);
	}

	fn render_dpad_buttons(&self, area: Rect, buf: &mut Buffer) {
		Block::bordered().render(area, buf);

		let up_button = self.create_button_ui(gilrs::Button::DPadUp, "Up");
		let right_button = self.create_button_ui(gilrs::Button::DPadRight, "Right");
		let down_button = self.create_button_ui(gilrs::Button::DPadDown, "Down");
		let left_button = self.create_button_ui(gilrs::Button::DPadLeft, "Left");

		let layers = Layout::vertical([Fill(1), Fill(1), Fill(1)]);
		let [top, mid, bot] = layers.areas(area);

		let top_layer = Layout::horizontal([Fill(1), Fill(1), Fill(1)]);
		let [_, up, _] = top_layer.areas(top);

		let mid_layer = Layout::horizontal([Fill(1), Fill(1), Fill(1)]);
		let [left, _, right] = mid_layer.areas(mid);

		let bot_layer = Layout::horizontal([Fill(1), Fill(1), Fill(1)]);
		let [_, down, _] = bot_layer.areas(bot);

		up_button.render(up, buf);
		right_button.render(right, buf);
		down_button.render(down, buf);
		left_button.render(left, buf);
	}

	fn render_menu_buttons(&self, area: Rect, buf: &mut Buffer) {
		let select_button = self.create_button_ui(gilrs::Button::Select, "Select");
		let start_button = self.create_button_ui(gilrs::Button::Start, "Start");

		let layers = Layout::horizontal([Fill(1), Fill(1), Fill(1), Fill(1), Fill(1)]);
		let [_, select_area, _, start_area, _] = layers.areas(area);

		let select_layer = Layout::vertical([Fill(1), Fill(1), Fill(5)]);
		let [_, select, _] = select_layer.areas(select_area);

		let start_layer = Layout::vertical([Fill(1), Fill(1), Fill(5)]);
		let [_, start, _] = start_layer.areas(start_area);

		select_button.render(select, buf);
		start_button.render(start, buf);
	}

	fn render_trigger_buttons(&self, area: Rect, buf: &mut Buffer) {
		let r1_button = self.create_button_ui(gilrs::Button::RightTrigger, "R1");
		let r2_button = self.create_button_ui(gilrs::Button::RightTrigger2, "R2");
		let l1_button = self.create_button_ui(gilrs::Button::LeftTrigger, "L1");
		let l2_button = self.create_button_ui(gilrs::Button::LeftTrigger2, "L2");

		let layers = Layout::horizontal([
			Fill(2),
			Fill(1),
			Fill(1),
			Fill(1),
			Fill(4),
			Fill(1),
			Fill(1),
			Fill(1),
			Fill(2),
		]);
		let [_, l2_area, _, l1_area, _, r1_area, _, r2_area, _] = layers.areas(area);

		let r1_layer = Layout::vertical([Fill(1), Fill(1), Fill(2)]);
		let [_, r1, _] = r1_layer.areas(r1_area);

		let r2_layer = Layout::vertical([Fill(1), Fill(4), Fill(2)]);
		let [_, r2, _] = r2_layer.areas(r2_area);

		let l1_layer = Layout::vertical([Fill(1), Fill(1), Fill(2)]);
		let [_, l1, _] = l1_layer.areas(l1_area);

		let l2_layer = Layout::vertical([Fill(1), Fill(4), Fill(2)]);
		let [_, l2, _] = l2_layer.areas(l2_area);

		r1_button.render(r1, buf);
		r2_button.render(r2, buf);
		l1_button.render(l1, buf);
		l2_button.render(l2, buf);

		//TODO: For PS4+ controllers doesn't have dedicated trigger push force
		// let gauge_percentage = match self.manager.active_gamepad() {
		// 	Ok(gamepad) => match gamepad.axis_data(gilrs::Axis::Unknown) {
		// 		Some(ad) => ad.value(),
		// 		None => 0.0,
		// 	},
		// 	Err(_) => 0.0,
		// };
		// let gauge_percentage = 0;

		// Gauge::default()
		// 	.gauge_style(Color::Green)
		// 	.bg(Color::DarkGray)
		// 	.ratio(gauge_percentage.into())
		// 	.render(r2, buf);
	}

	fn render_joystick_areas(&self, area: Rect, buf: &mut Buffer) {
		let layout = Layout::horizontal([Fill(4), Fill(4), Fill(2), Fill(4), Fill(4)]);
		let [_, left, _, right, _] = layout.areas(area);

		self.create_joystick_ui(gilrs::Axis::LeftStickX, gilrs::Axis::LeftStickY)
			.render(left, buf);

		self.create_joystick_ui(gilrs::Axis::RightStickX, gilrs::Axis::RightStickY)
			.render(right, buf);
	}

	fn create_button_ui(&self, btn: gilrs::Button, title: &'static str) -> Block {
		let color = match self.manager.active_gamepad() {
			Ok(gamepad) => match gamepad.is_pressed(btn) {
				true => Color::Green,
				false => Color::DarkGray,
			},
			Err(_) => Color::DarkGray,
		};

		Block::new()
			.title_top(title)
			.style(Style::default().bg(color))
	}

	fn create_joystick_ui(&self, axis_x: gilrs::Axis, axis_y: gilrs::Axis) -> impl Widget + '_ {
		let (x, y) = match self.manager.active_gamepad() {
			Ok(gp) => {
				let left_axis = gp.axis_data(axis_x).map_or(0.0, |d| d.value()); // 160.0
				let right_axis = gp.axis_data(axis_y).map_or(0.0, |d| d.value()); // 140.0

				(left_axis as f64, right_axis as f64)
			}
			Err(_) => (0.0_f64, 0.0_f64),
		};

		let color = match (x, y) {
			(a, b)
				if a < REST_DEADZONE
					&& b < REST_DEADZONE
					&& a > -REST_DEADZONE
					&& b > -REST_DEADZONE =>
			{
				Color::DarkGray
			}
			(a, b)
				if a > MAX_DEADZONE
					|| b > MAX_DEADZONE
					|| a < -MAX_DEADZONE
					|| b < -MAX_DEADZONE =>
			{
				Color::Green
			}
			(a, b)
				if a < MIN_DEADZONE
					|| b < MIN_DEADZONE
					|| a > -MIN_DEADZONE
					|| b > -MIN_DEADZONE =>
			{
				Color::Red
			}
			(_, _) => Color::DarkGray,
		};

		let (x, y) = (x * 160.0, y * 140.0);

		let joystick = Rectangle {
			x: x - 20.0,
			y: y - 40.0,
			width: 40.0,
			height: 60.0,
			color,
		};

		Canvas::default()
			.block(
				Block::bordered()
					.border_type(BorderType::QuadrantInside)
					.border_style(Color::DarkGray),
			)
			.marker(Marker::HalfBlock)
			.paint(move |ctx| ctx.draw(&joystick))
			.x_bounds([-180.0, 180.0])
			.y_bounds([-180.0, 180.0])
	}
}

impl Widget for &App {
	fn render(self, area: Rect, buf: &mut Buffer) {
		let layout = Layout::vertical([Length(3), Percentage(100)]);
		let [header_area, inner_area] = layout.areas(area);

		let gamepad_layout = Layout::vertical([Fill(8), Fill(10), Fill(2), Fill(10), Fill(2)]);
		let [triggers, buttons, _, joysticks, _] = gamepad_layout.areas(inner_area);

		let buttons_layout = Layout::horizontal([Fill(1), Fill(10), Fill(10), Fill(10), Fill(1)]);
		let [_, left_buttons, mid_buttons, right_buttons, _] = buttons_layout.areas(buttons);

		let mut title_label = Block::bordered();
		title_label = match self.manager.active_gamepad() {
			Ok(gp) => title_label
				.title(gp.name().to_string())
				.title_style(Color::Rgb(255, 165, 35)),
			Err(_) => title_label.title("No active device".to_string()),
		};
		title_label.render(header_area, buf);

		self.render_dpad_buttons(left_buttons, buf);
		self.render_menu_buttons(mid_buttons, buf);
		self.render_action_buttons(right_buttons, buf);
		self.render_trigger_buttons(triggers, buf);
		self.render_joystick_areas(joysticks, buf);
	}
}
